const getAdvancedHTTPProperties = require('./http-properties');

/**
 * This factory is responsible for creation of metadata for axway-requester.
 * axway-requester use it for communicating properly with third party services.
 * @param {string} uri the method uri
 * @returns {object} object that contains requester metadata and function that adds more metadata to it.
 */
module.exports = (uri) => {
	const metadata = {
		uri,
		methods: {}
	};

	return { metadata, addMethodMetadata };

	function addMethodMetadata ({ customOperationID, mpath, httpMethod, securityDefinitions, method, swagger }) {
		metadata.methods[customOperationID] = {
			path: mpath,
			httpMethod,
			security: method.security,
			securityDefinitions,
			parameters: constructParameters(method, swagger)
		};
	}

	function constructParameters(method, swagger) {
		const orderedParameters = {
			query: [],
			header: [],
			path: [],
			body: [],
			formData: [],
			// This holds the default values for some of the advanced HTTP parameters.
			// Usually we can hardcode the default values but in this case the default values are dynamic so we have to pass them down.
			defaultValues: []
		};
		const parameters = method.parameters || [];
		parameters.reduce((acc, param) => {
			const place = param.in;
			acc[place].push(param);
			return acc;
		}, orderedParameters);

		const hasBodyParam = orderedParameters.body.length > 0;
		const advancedHTTPProperties = getAdvancedHTTPProperties(swagger, method, hasBodyParam);
		advancedHTTPProperties.header.forEach((param) => {
			// Add it as parameter
			orderedParameters.header.push({ name: param.name });
			// Store its default value to use it in case it is not provided
			orderedParameters.defaultValues.push({
				in: 'header',
				name: param.name,
				defaultValue: param.value.default
			});
		});

		return orderedParameters;
	}
};
